
// called just before sensible.Application.start()
sensible.fxos.Application.prototype.onBeforeStart = function (inCallback)
{
	console.log ("fxos.Application.onBeforeStart()");

	// save away the callback
	// as we'll need it for when the discover command returns
	this.discoverCallback = inCallback;

	console.log ("making socket");
	
	try
	{
		// ok we need to find the Little Sky and set various things
		// before we allow Sensible to continue
		this.socket = new UDPSocket ();
	}
	catch (inError)
	{
		console.error ("error opening socket");
		console.error (inError);
	}
	
	var	self = this;
	
	this.socket.opened.then
	(
		function ()
		{
			console.log ("socket is open");
			
			self.socket.addEventListener
			(
				"message",
				function (inMessage)
				{
					self.onMessage (inMessage);
				}
			);

			self.findLittleSky ();
		}
	);
}

// this assumes we are on a class C
sensible.fxos.Application.prototype.findLittleSky = function ()
{
	if (navigator.mozWifiManager && navigator.mozWifiManager.connectionInformation
		&& navigator.mozWifiManager.connectionInformation.ipAddress)
	{
		var	localAddress = navigator.mozWifiManager.connectionInformation.ipAddress;
		var	localAddressElements = localAddress.split (".");
		localAddressElements.length = 3;
		var	localSubnet = localAddressElements.join (".");
		
		var	self = this;
		var	host = 1;
		
		// every 250ms, try an address
		var	interval = setInterval
		(
			function ()
			{
				if (self.littleSkyAddress)
				{
					// discover callback was called, we found it
					clearInterval (interval);
					
					var	addressElement = document.getElementById ("little_sky_address");
					
					if (addressElement)
					{
						addressElement.innerHTML = "Found " + self.littleSkyNickname + " at " + self.littleSkyAddress;
					}

					var	proxyElement = document.getElementById ("proxy_address");
					
					if (proxyElement)
					{
						proxyElement.innerHTML = "Proxy at http://" + self.config.hostname + ":" + self.config.port;
					}
				}
				else
				{
					var	remoteAddress = localSubnet + "." + host;
					console.log ("trying " + remoteAddress);
					
					self.socket.send (self.str2ab ("!discover;"), remoteAddress, 50007);
					
					host++;
					
					if (host == 255)
					{
						console.error ("reached 254, giving up");
						clearInterval (interval);
					}
				}
			},
			250
		);
	}
	else
	{
		console.error ("couldn't get ip address from mozWifiManager");
	}
}

// called when our socket gets a packet
sensible.fxos.Application.prototype.onMessage = function (inMessage)
{
	var	messageString = this.ab2str (inMessage.data);
	// var	messageString = inMessage.data;

	console.log ("Application.onMessage(" + messageString + ")");
	console.log (typeof messageString);
	
	if (messageString.charAt (0) == "!")
	{
		console.log ("little sky message received");
		console.log (messageString);
		
		// strip the opening ! and trailing ;
		messageString = messageString.slice (1, messageString.length - 1);
		
		var	messageElements = messageString.split (",");
		var	messageStatus = messageElements.splice (0, 1) [0];
		
		if (messageStatus == "OK")
		{
			var	messageOpcode = messageElements.splice (0, 1) [0];
			messageOpcode = messageOpcode.toLowerCase ();
			
			// find a handler!
			var	handlerName = "ls_message_" + messageOpcode;
			
			if (typeof (this [handlerName]) == "function")
			{
				this [handlerName].call (this, messageElements);
			}
			else
			{
				console.error ("couldn't find handler " + handlerName);
			}
		}
		else
		{
			console.error ("message status error");
		}
	}
	else
	{
		console.error ("non-LS message received");
		console.error (messageString);
	}
}

// called just after sensible.Application.start()
sensible.fxos.Application.prototype.onAfterStart = function (inCallback)
{
	console.log ("fxos.Application.onAfterStart()");
	
	inCallback ();
}

// LS MESSAGE HANDLERS

sensible.fxos.Application.prototype.ls_message_discover = function (inParameters)
{
	console.log ("ls_message_discover()");
	
	this.littleSkyAddress = inParameters [0];
	this.littleSkyNickname = inParameters [1];
	
	if (this.littleSkyNickname == "<unnamed>")
	{
		this.littleSkyNickname = "Little Sky";
	}

	// copy the nickname to the config
	// so that's what shows up in MDNS searches
	this.config.name = this.littleSkyNickname;
	
	console.log ("found LS name of " + this.config.name);
	
	// if we have a callback cached, call it to allow app startup
	if (this.discoverCallback)
	{
		console.log ("calling discover callback");
		this.discoverCallback.call (this);
		this.discoverCallback = null;
	}
}

// OVERRIDES

// we override all property set calls
sensible.fxos.Application.prototype.properties_set = function (inRequest, inCallback)
{
	console.log ("Application.properties_set()");
	
	for (var param in inRequest.parameters)
	{
		if (typeof param == "string")
		{
			if (this.propertiesByKey [param])
			{
				var	value = inRequest.parameters [param];

				if (typeof (value) == "string")
				{
					if (param == "brightness")
					{
						var	brightness = parseInt (value);
						console.log ("setting brightness to " + brightness);
					
						this.sendPacket ("!setbr," + brightness + ";");
					}
					else
					if (param == "br_up")
					{
						this.sendPacket ("!br_up;");
					}
					else
					if (param == "br_down")
					{
						this.sendPacket ("!br_down;");
					}
					if (param == "next_playlist")
					{
						this.sendPacket ("!next_playlist;");
					}
					else
					{
						console.log ("no handler for registered param: " + param);
					}
				}
			}
		}
		
		console.log (param + " = " + inRequest.parameters [param]);
	}
	
	var	response = new Object ();
	response.type = "json";
	response.object = null;
	
	inCallback (response);
}

// LITTLESKY APIS

sensible.fxos.Application.prototype.sendPacket = function (inMessageText)
{
	console.log ("sendPacket(" + inMessageText + ")");
	
	if (this.littleSkyAddress)
	{
		var	messageBuffer = this.str2ab (inMessageText);
		this.socket.send (messageBuffer, this.littleSkyAddress, 50007);
	}
	else
	{
		console.error ("request to send packet with no little sky address set");
	}
}

// PRIVATE

// this is quick and dirty because the socket API is also
sensible.fxos.Application.prototype.str2ab = function (inString)
{
	var	buffer = new ArrayBuffer (inString.length);
	var	view = new Uint8Array (buffer);
	
	for (var i = 0; i < inString.length; i++)
	{
		view [i] = inString.charCodeAt (i);
	}
	
	return buffer;
}

// this is quick and dirty because the socket API is also
sensible.fxos.Application.prototype.ab2str = function (inBuffer)
{
	var	string = "";
	var	view = new Uint8Array (inBuffer);
	
	for (var i = 0; i < inBuffer.byteLength; i++)
	{
		string += String.fromCharCode (view [i]);
	}
	
	return string;
}

